/**************************************************************************/
/*                                                                        */
/*  Thibaut Balabonski, Sylvain Conchon, Jean-Christophe Filliâtre,       */
/*  Kim Nguyen, Laurent Sartre                                            */
/*                                                                        */
/*  Informatique - MP2I/MPI - CPGE 1re et 2e années.                      */
/*  Cours et exercices corrigés. Éditions Ellipses, 2022.                 */
/*                                                                        */
/*  https://www.informatique-mpi.fr/                                      */
/*                                                                        */
/**************************************************************************/

// composantes fortement connexes d'un graphe orienté
// algorithme de Kosaraju--Sharir

#include "prelude.h"
#include "digraph.h"
#include "kosaraju_sharir.h"

// attention : ce code n'est pas réentrant à cause de nc
//   (bien sûr, on pourrait passer un pointeur sur nc, mais le code
//   deviendrait plus complexe à suivre)

int nc; // nombre total de composantes

void dfs(digraph *g, int v, bool visited[], int num[]) {
  if (visited[v]) return;
  visited[v] = true;
  num[v] = nc;
  for (list *l = digraph_succ(g, v); l != NULL; l = l->next)
    dfs(g, l->value, visited, num);
}

int kosaraju_sharir(digraph *g, int num[]) {
  int n = digraph_size(g);
  nc = 0;
  bool *visited = calloc(n, sizeof(bool));
  for (int i = 0; i < n; i++) visited[i] = false;
  int *order = calloc(n, sizeof(int));
  digraph_postorder(digraph_mirror(g), order);
  for (int i = 0; i < n; i++) {
    int v = order[i];
    if (visited[v]) continue;
    dfs(g, v, visited, num);
    nc++;
  }
  return nc;
}
